#!/usr/bin/env python
"""
DHCP* DNS Record Generator (ddrgen), version 1.1.0
Copyright 2005-2006 Michael Farrell

http://lz129.concreteairship.com/projects/ddrgen/

This software scans the local network for open SMB/CIFS (Windows File and
Printer Sharing) servers/shares, and maps them to DNS records.  It will
additionally (optionally) write statistics to an XML file so that you can 
display that data on a web page.

Syntax:
 ./ddrgen.py (domain) (network) (mask) (ns ip) (update freq) (fwd zone file) [rev zone file] [report file] [extra options]

       (domain) - The domain of the network you want records for.

      (network) - The network the records are for.  For example, if your
                  network is 192.168.0.1 -> 192.168.0.254, you would write
                  192.168.0.0 here.

         (mask) - The number of octects the network spans.  Only "whole"
                  ones should be used, so only 8, 16 or 24 may appear
                  here.  It's untested with others, so beware.  With the
                  above example, you would use 24.

        (ns ip) - The nameserver's IP.

  (update freq) - The frequency, in seconds, that you are running this
                  script (or how often clients should look for updates)

(fwd zone file) - The file to write the forward records to.  If this file is
                  named 'data', it is assumed you wish to output in a djbdns
                  /tinydns format, so the reverse file parameter isn't needed

(rev zone file) - The file to write the reverse records to.

  [report file] - (optional) A file to write a XML file with a report of
                  what machines are on the network.
                  
[extra options] - (optional) Valid options are:
                  nogames - Don't scan for games.  This makes things faster.

Please note that the forward and reverse zone files will be totally
overwritten on update.

"""

import sys, string, os, time

try:
	import ddrghelpers
except:
	print "FATAL: Couldn't import ddrghelpers.  Is it in your Python search path?"
	sys.exit(1)

# format is: "Original: (Real Name, Sastisfiable Match)"
osList = {
	# This is the "x64" version of Windows XP.  It uses a Win2003 kernel
	# so therefore it is Windows 5.2, and only goes to SP1.
	"Windows XP 3790 Service Pack 1": ("Windows XP Pro/amd64 (5.2.3790, SP1)", 1),
	
	
	# TODO: Windows 2003 for amd64.  This is only the i386 version.
	"Windows Server 2003 3790": ("Windows 2003 Server (5.2.3790, SP1?)", 1),
	# Some 
	"Windows Server 2003 3790 Service Pack 1": ("Windows 2003 Server (5.2.3790, SP1)", 1),
	
	# These are all vanilla i386 versions.
	# TODO: Confirm WinME, Win98, Win95.  Also get WinNT string.
	"Windows 5.1": ("Windows XP/i386", 1),
	"Windows 5.0": ("Windows 2000", 1),
	"Windows 4.9": ("Windows ME", 1), # needs confirmation
	"Windows 4.1": ("Windows 98", 1),
	"Windows 4.0": ("Windows 95", 1), # may conflict with NT 4?
	
	"Unix": ("Unix", 0), # Probably samba, unknown OS.  Mac OS X also reports as this.
	"": ""}

# This is to try and guess the OS from the patches applied to Samba that are nicely reported to us.
serverToOsList = {
	# Debian GNU/Linux.  Please note I'm tracking Sarge here.
	"Samba 3.0.14a-Debian": "Debian GNU/Linux", # Sarge at release; 20050602.
	"Samba 3.0.14-Debian": "Debian GNU/Linux", # Sarge
	"Samba 3.0.10-Debian": "Debian GNU/Linux", # Sarge as at 20050430
	
	# Ubuntu Linux
	"Samba 3.0.20b-Ubuntu": "Ubuntu Linux (6.04 \"Dapper\")", # Ubuntu Dapper (pre-release), at 20060304.
	"Samba 3.0.14a-Ubuntu": "Ubuntu Linux (5.10 \"Breezy\")", # Ubuntu Breezy (release), at 20051126.
	"Samba 3.0.10-Ubuntu": "Ubuntu Linux (5.04 \"Hoary\")", # Ubuntu Hoary as at 20050430
	
	# Clarkconnect
	"Samba 3.0.10-3.cc": "Clarkconnect Linux",
	
	# Generic ones.  Used by Arch and others.
	# Mac OS X doesn't report a specific version, however you should be
	# able to see the outspoken Mac and OSx86 users at a LAN Party.
	# Generally people will be hanging around their rigs...
	"Samba 3.0.14a": "GENERIC",
	"Samba 3.0.14": "GENERIC",
	"Samba 3.0.10": "GENERIC",
	"": ""}

# check our inputs.
if len(sys.argv) < 7:
	print "FATAL: Oops, you need some more parameters..."
	print __doc__	
	sys.exit(1)
	
domain = sys.argv[1]
network = sys.argv[2]
subnet = int(sys.argv[3])
nsIp = sys.argv[4]
updateFreq = int(sys.argv[5])
fwdFile = sys.argv[6]
djbdns = 0

if fwdFile[-4:] == 'data': # probably djbdns/tinydns.
	djbdns = 1
	if len(sys.argv) >= 8:
		statsFile = sys.argv[7]
		genStats = 1
	else:
		genStats = 0	
else:
	revFile = sys.argv[7]

	if len(sys.argv) >= 9:
		statsFile = sys.argv[8]
		genStats = 1
	else:
		genStats = 0

scanForGames = 1
for x in sys.argv[7:]:
	if x == "nogames":
		scanForGames = 0

ddrghelpers.selfTest(djbdns, scanForGames)

if scanForGames == 1:
	# try to import game modules
	# some installations may be incorrect, or stripped to save space, so we
	# handle this with GRACE.
	try:
		import ut2004, q3a
	except:
		print "WARNING: There was a problem trying to load the game scanning support modules.  This probably means that 'q3a.py' and 'ut2004.py' aren't in your Python search path, or one of their dependancies failed to load correctly."
		print "This program will continue to run, however the 'nogames' option is assumed, so no game servers will be reported."
		scanForGames = 0

if (subnet > 31) or (subnet < 8):
	print "FATAL: Invalid subnet, it should be between 8 and 31 (ideally 8, 16 or 24)."
	sys.exit(1)

# check network address
ddrghelpers.checkNetworkAddress(network, subnet)

# Check the NS IP
ddrghelpers.checkIpAddress(nsIp)

print "Summary of my orders are:"
print " - Scan " + network + "/" + str(subnet) + " for SMB servers"
print " - The expire time of data will be in " + str(updateFreq) + " seconds"
if djbdns == 0:
	print " - Create a forward DNS in " + fwdFile
	print " - Create a reverse DNS in " + revFile
else:
	print " - Create a djbdns-format DNS in " + fwdFile
if genStats == 1:
	print " - Generate stats in " + statsFile
if scanForGames == 1:
	print " - Scan for game servers"
else:
	print " - WILL NOT scan for game servers"
print "Starting now..."

print "Scanning for hosts..."
# make a list of up SMB hosts in ddrgen.tmp. add -P0 on nmap to find windows xp firewall'd boxen
os.system("nmap -n -P0 -p 139 \'" + network + "/" + str(subnet) + "\' | grep -B2 \'cp o\' | grep on | awk \'{print $4}\' | grep ':' | sed s/\://g > ddrgen.tmp")

# ok, now read that file into memory.
fh = open("ddrgen.tmp", "r")
ipAddrs = fh.readlines()
fh.close()

print "Searching SMB server shares..."

# resolve IPs to nmb-names (or wins-names, whatever floats your boat)
servers = []
for ip in ipAddrs:
	ip = ip.replace("\n", "")
	os.system("nmblookup -A " + ip + " | grep \'<00>\' | grep -v GROUP | awk \'{print $1}\' > ddrgen.tmp")
	fh = open("ddrgen.tmp", "r")
	servers.append((ip, fh.readline().replace("\n", "").lower()))
	fh.close()

# try to open the fwd and rev zone files
try:
	fwdfh = open(fwdFile, "w")
	if djbdns == 0:
		revfh = open(revFile, "w")
except:
	print "FATAL: Error opening zones files for writing."

# write headers...

print "Writing DNS entries..."
if djbdns == 1:
	# djbdns/tinydns version
	ddrghelpers.iterateDjbDnsRecords(fwdfh, servers, domain, nsIp, time, updateFreq)
else:
	ddrghelpers.iterateBindRecords(fwdfh, revfh, servers, domain, nsIp, subnet, time, updateFreq)
	revfh.close()

fwdfh.close()

# statistics generation...

if genStats == 1:
	if scanForGames == 1:
		print "Generating statistics file, and scanning for shares and game servers.  If firewalls drop rather than reject packets (the 'Steve Gibson' way, like in Windows XP/2003 Firewall, and ZoneAlarm), scanning for games can take a VERY LONG (sometimes infinite) time."
		print "If this takes too long, then you probably have a host on the network like that.  Please kill this python process, and LART the user responsible."
		# Tip to those reading this source code:
		# ping -f works well, particularly if they have popup alerts
		# enabled. :)
		# Of course, you may be booted out of the LAN if you get
		# caught, and some people call it "hacking".
	else:
		print "Generating statistics file, and scanning for shares..."

	try:
		statsfh = open(statsFile, "w")
	except:
		print "FATAL: Error opening stats file for writing."
		sys.exit(1)

	statsfh.write("<?xml version=\"1.0\"?>\n<stats>\n")
	statsfh.write("<updated>" + time.strftime("%c") + "</updated>")
	# Generate some statistics.
	for ip, name in servers:
		os.system("smbclient -gNL " + ip + " > ddrgen.tmp 2>&1")
		fh = open("ddrgen.tmp", "r")
		serverinfo = fh.readlines()
		fh.close()
		servstring = serverinfo[0].split("]")
		
		try:
			servOS = servstring[1][5:]
		except:
			servOS = "Unknown"
		
		try:
			servServer = servstring[2][9:]
		except:
			servServer = "Unknown"
				
		osSuccess = 0
		try:
			osSuccess = osList[servOS][1]
			servOS = osList[servOS][0]
		except:
			print "WARNING: Unrecogized OS '" + servOS + "' on '" + name + "' (" + ip + ") . If you know it's OS, please contact me so more OSes can be put in the database."
		
		# try to guess OS from server name
		if osSuccess != 1:
			try:
				if serverToOsList[servServer] != "GENERIC":
					servOS = serverToOsList[servServer]
			except:
				print "WARNING: Unknown server name and unsatisfied with reported OS information (" + servOS + ").  The server name was '" + servServer + "', running on '" + name + "' (" + ip + ") . If you know it's OS and/or server software, please contact me so more OSes and servers can be put in the database."
		diskinfo = "<shares>"
		for disk in serverinfo:
			disk = disk[:-1] # trim nl
			if (disk[:3] != "Ano") and (disk[:3] != "Dom"): # skip more messages...
				diskdata = disk.split("|")
				if diskdata[0] == "Disk":
					# append disk data.
					diskinfo = diskinfo + "<share><type>smb</type><name>" + diskdata[1] + "</name><comment>" + diskdata[2] + "</comment></share>"
		diskinfo = diskinfo + "</shares>"
		
		# GAMES SCANNING
		if scanForGames == 1:
			# look for ut200x
			ut2004stats = ut2004.retrieveStatus(ip)
			ut2004data = ""
			if ut2004stats != 0:
				ut2004data = "<game>"
				# there is a ut2004 server, get data!
				for nameU, valueU in ut2004stats[0].iteritems():
					ut2004data = ut2004data + "<" + nameU.replace(" ", "") + ">" + str(valueU) + "</" + nameU + ">"
				for nameU, valueU in ut2004stats[1].iteritems():
					ut2004data = ut2004data + "<" + nameU.replace(" ", "") + ">" + str(valueU) + "</" + nameU + ">"
				ut2004data = ut2004data + "</game>"

			# look for q3a
			tempdata = ""
			tempstats = q3a.retrieveStatus(ip)
			if tempstats != 0:
				tempdata = "<game>"
				for nameU, valueU in tempstats.iteritems():
					tempdata = tempdata + "<" + nameU.replace(" ", "") + ">" + str(valueU) + "</" + nameU + ">"
				q3adata = tempdata + "</game>"		
			else:
				q3adata = ""

			# look for Call of Duty
			tempdata = ""
			tempstats = q3a.retrieveStatus(ip, 28960)
			if tempstats != 0:
				tempdata = "<game>"
				for nameU, valueU in tempstats.iteritems():
					tempdata = tempdata + "<" + nameU.replace(" ", "") + ">" + str(valueU) + "</" + nameU + ">"
				coddata = tempdata + "</game>"		
			else:
				coddata = ""

		statsfh.write("<computer><ip>" + ip + "</ip><name>" + name + "</name><os>" + servOS + "</os><server>" + servServer + "</server>" + diskinfo)
		if (scanForGames == 1):
			if (len(ut2004data + q3adata + coddata) > 0):
				statsfh.write("<games>" + ut2004data + q3adata + coddata + "</games>")
		statsfh.write("</computer>\n")

	statsfh.write("</stats>")
	statsfh.close()

# remove our tempfile.
os.remove("ddrgen.tmp")

# report that we are done.
print "All done!"
