#!/bin/false
"""ddrghelpers.py: Modules with reusable code in ddrgen.
Copyright 2005-2006 Michael Farrell.
"""

import struct

# define ourself some extra exceptions
class ImproperIpFourException(Exception): pass
class ImproperNetworkAddressException(Exception): pass
class CidrOutOfRangeException(Exception): pass

# module functions.
def selfTest(djbdns, scanForGames):
	"""Does a quick check on the system to see if there are any issues with the environment.

	'djbdns' says wether djbdns is used or not.  Similarly with 'scanForGames'.  This determines the level of testing done."""
	
	if djbdns == 1:
		try:
			struct.pack(">Q", long(2**62))
		except:
			print "WARNING: This installation of Python does not support 64-bit 'long long' datatypes.  Please ensure you are running Python 2.2 or later, and that your host system and compiler support this.  ddrgen will attempt to gracefully 'fall back', however errors may occur with the software."
	
	if scanForGames == 1:
		try:
			from twisted.internet.protocol import DatagramProtocol
			from twisted.internet import reactor
		except:
			print "WARNING: Couldn't load twisted networking subsystem.  Please make sure it is installed, and in your Python module path.  You will not be able to scan for game servers without it."
	
	print "Self-test finished.  No fatal errors."
			

def checkIpAddress(ip):
	"""Checks the IP 'ip' to make sure it is correct and well-formed.  Throws ImproperIpFourException on error, returns IP in list if correct."""
	ipArr = ip.split(".")
	
	if len(ipArr) != 4:
		raise ImproperIpFourException, "Network address must have four 'numbers'."
	
	# check the network address and make the bits all integer.
	y = 0
	for x in ipArr:
		x = int(x)
		ipArr[y] = x
		y = y + 1
		if (x < 0) or (x > 255):
			raise ImproperIpFourException, "Octet outside range of 0-255."
	
	return ipArr

def checkNetworkAddress(ip, mask):
	"""Does a basic check to see that a network address is well formed."""
	networkArr = checkIpAddress(ip)
	
	mask = int(mask)
	if (mask > 24) or (mask < 8):
		raise CidrOutOfRangeException, "CIDR 'mask' must be in the range 8 to 24."
	
	if ((networkArr[3] != 0) and (mask <= 24)) or ((networkArr[2] != 0) and (mask <= 16)) or ((networkArr[1] != 0) and (mask <= 8)):
		raise ImproperNetworkAddressException, "Network address isn't correct for the size specified.  Do octets equal zero where needed?"
	
	return networkArr

def isBadNetworkName(name):
	"""Checks if 'name' is "bad".  Particularly needed for some paranoid Primary Domain Controllers, that refuse to state their name in some configurations.
	Returns 1 if bad, 0 otherwise."""
	name = name.lower()
	if (name == "ns") or (name == "ddrlan") or (len(name) <= 1):
		return 1
	else:
		return 0

def makeTai64Timestamp(time):
	"""Makes a TAI-64 timestap from the time object "time".  See http://cr.yp.to/libtai/tai64.html#tai64 for info."""
	# create a decimal tai64 value.  unix time + 2^62.
	taidec = long(time.time()) + long(2**62)
	
	# pack 64-bit long long into string, big endian.
	taihex = struct.pack(">Q", taidec)
	
	
	# convert 8 byte string into 16 bytes base16 string (hex), and return.
	return ''.join(["%02X"%ord(t) for t in taihex])

def iterateDjbDnsRecords(fwdfh, servers, domain, nsIp, time, updateFreq):
	"""Makes some bulk djbdns records.  fwdfh is a file handle for 'data', servers is a list of servers found, domain is the domain of the network."""
	nsfound = 0

	# 'gracefully' fall back if there's no long long (64-bit long) support.
	try:
		timestamp = ":" + makeTai64Timestamp(time)
	except:
		timestamp = ""
	
	fwdfh.write("." + domain + ":" + nsIp + ":ddrlan:" + str(updateFreq) + timestamp + "\n")
	for ip, name in servers:
		if isBadNetworkName(name) == 0:
			fwdfh.write("=" + name + "." + domain + ":" + ip + "\n")
			
			# The nameserver has an SMB server.
			if ip == nsIp:
				nsfound = 1
				

		
	fwdfh.write("+ns." + domain + ":" + nsIp + "\n")

def iterateBindRecords(fwdfh, revfh, servers, domain, nsIp, subnet, time, updateFreq):
	"""Makes some bulk BIND-style records.  'fwdfh' and 'revfh' are filehandles, 'time' is a "time" object, and all the rest is self explanitory."""
	fwdfh.write("$ttl " + str(updateFreq) + """
""" +domain + ".	IN	SOA	ns." + domain + ".	noemail." + domain + """. (
	""" + str(int(time.time())) + """
	""" + str(updateFreq) + """
	""" + str(updateFreq) + """
	""" + str(updateFreq * 5) + """
	5m )

""" + domain + """.	IN	NS	ns.""" + domain + """.
ns.""" + domain + """.	IN	A	""" + nsIp + """
""")

	if subnet < 16:
		arpadomain = str(networkArr[0]) + ".in-addr.arpa"
	else:
		if subnet < 24:
			arpadomain = str(networkArr[1]) + "." + str(networkArr[0]) + ".in-addr.arpa"
		else: # < 24... or waddever.
			arpadomain = str(networkArr[2]) + "." + str(networkArr[1]) + "." + str(networkArr[0]) + ".in-addr.arpa"

	revfh.write("$ttl " + str(updateFreq) + """
""" +arpadomain + ".	IN	SOA	ns." + domain + ".	noemail." + domain + """. (
	""" + str(int(time.time())) + """
	""" + str(updateFreq) + """
	""" + str(updateFreq) + """
	""" + str(updateFreq * 5) + """
	5m )

""" + arpadomain + """.	IN	NS	ns.""" + domain + """.
""")

	# write out servers found
	for ip, name in servers:
		if ddrghelpers.isBadNetworkName(name) == 0:
			fwdfh.write(name + "." + domain + ".	IN	A	" + ip + "\n")
			ipArr = ip.split(".")
			revfh.write(ipArr[3] + "." + ipArr[2] + "." + ipArr[1] + "." + ipArr[0] + ".in-addr.arpa.	IN	PTR	" + name + "." + domain + "." + "\n")

